// PersistentSet.java, created Thu Apr  8 02:32:18 1999 by cananian
// Copyright (C) 1999 C. Scott Ananian <cananian@alumni.princeton.edu>
// Licensed under the terms of the GNU GPL; see COPYING for details.
package net.cscott.jutil;

import java.util.AbstractSet;
import java.util.Comparator;
import java.util.Iterator;
import java.util.Set;
/**
 * {@link PersistentSet} implements a persistent set, based on a
 * persistent randomized treap.  Unlike the {@link Set}s returned
 * by {@link PersistentSetFactory}, {@link PersistentSet}
 * does not implement that standard {@link java.util.Set} API
 * but instead exposes the underlying functional operations.
 * 
 * @author  C. Scott Ananian <cananian@alumni.princeton.edu>
 * @version $Id: PersistentSet.java,v 1.8 2006-10-30 19:58:06 cananian Exp $
 */
public class PersistentSet<T>  implements Iterable<T> {
    final Node<T> root;
    final Comparator<T> c;
    final Node.Allocator<T> allocator;

    /** Creates an empty {@link PersistentSet} whose member objects
     *  will all implement {@link java.lang.Comparable}. Note that
     *  good hashcode implementations for the member objects are still
     *  required. */
    public PersistentSet() {
	// cast below is safe iff T implements Comparable.
	this((Comparator<T>)Default.comparator); // safe iff T is Comparable
    }
    /** Creates an empty {@link PersistentSet} whose member objects
     *  are ordered by the given {@link Comparator}.  Note that
     *  good hashcode implementations for the member objects are still
     *  required.
     */
    public PersistentSet(Comparator<T> c) {
	this(null, c, new Node.Allocator<T>());
    }
    /** Creates a {@link PersistentSet} from a root {@link Node}
     *  and a {@link Comparator}.*/
    private PersistentSet(Node<T> root, Comparator<T> c, Node.Allocator<T> a) {
	this.root = root; this.c = c; this.allocator = a;
    }

    /** Determines if this {@link PersistentSet} has any members. */
    public boolean isEmpty() { return (root==null); }

    /** Count the number of elements in this {@link PersistentSet}. */
    public int size() { return (root==null)?0:root.size; }

    /** Creates and returns a new {@link PersistentSet} identical to
     *  this one, except it contains <code>element</code>. */
    public PersistentSet<T> add(T element) {
	Node<T> new_root =
	    PersistentTreeNode.put(this.root, this.c, element, element,
				   allocator);
	return (this.root == new_root) ? this :
	    new PersistentSet<T>(new_root, c, allocator);
    }
    /** Determines if the given element belongs to this set. */
    public boolean contains(T element) {
	return (PersistentTreeNode.get(this.root, this.c, element)!=null);
    }

    /** Make a new {@link PersistentSet} identical to this one,
     *  except that it does not contain <code>element</code>. */
    public PersistentSet<T> remove(T element) {
	Node<T> new_root = 
	    PersistentTreeNode.remove(this.root, this.c, element, allocator);
	return (this.root == new_root) ? this :
	    new PersistentSet<T>(new_root, c, allocator);
    }
    /** Add all the items in the given set to this set. */
    public PersistentSet<T> addAll(PersistentSet<T> set) {
	if (this.c.equals(set.c) && this.allocator==set.allocator) {
	    Node<T> new_root =
		Node.putAll(this.root, set.root, this.c, allocator);
	    return (new_root==this.root) ? this :
		(new_root==set.root) ? set :
		new PersistentSet<T>(new_root, c, allocator);
	}
	// slow way
	if (set.size() > this.size()) return set.addAll(this);
	Node<T> new_root = this.root;
	for (Iterator<Node<T>> it=Node.iterator(set.root); it.hasNext(); ) {
	    Node<T> n = it.next();
	    new_root = Node.put(new_root, this.c, n.key, n.getValue(),
				allocator);
	}
	return (this.root == new_root) ? this :
	    new PersistentSet<T>(new_root, c, allocator);
    }

    /** Cloning takes constant time, regardless of the size of the set. */
    public PersistentSet<T> clone() {
	return new PersistentSet<T>(root, c, allocator);
    }
   
    /** Human-readable representation of the set. */
    public String toString() { return asSet().toString(); }
    
    /** Unmodifiable iterator. */
    public Iterator<T> iterator() {
        final Iterator<Node<T>> it = Node.iterator(root);
        return new UnmodifiableIterator<T>() {
            @Override
            public boolean hasNext() { return it.hasNext(); }
            @Override
            public T next() { return it.next().key; }
        };
    }
    
    /*---------------------------------------------------------------*/
    /** {@link java.util.Collection}s view of the set. */
    public Set<T> asSet() {
	return new AbstractSet<T>() {
	    // constant-time clone.
	    public Set<T> clone() { return PersistentSet.this.asSet(); }
	    public boolean contains(Object o) {
		// oops, not safe if we pass a non-T o in here!
		return PersistentSet.this.contains((T)o);
	    }
	    public boolean isEmpty() {
		return PersistentSet.this.isEmpty();
	    }
	    public int size() {
		return PersistentSet.this.size();
	    }
	    public Iterator<T> iterator() {
                final Iterator<Node<T>> it = Node.iterator(root);
		return new Iterator<T>() {
		    Node<T> last=null;
		    public boolean hasNext() { return it.hasNext(); }
		    public T next() {
			last = it.next();
			return last.key;
		    }
		    public void remove() {
			if (last==null)
			    throw new IllegalStateException();
			// XXX doesn't update the Set view
			// changes behind the iterator's back don't
			// matter, because everything's persistent!
			PersistentSet.this.remove(last.getKey());
			last=null;
		    }
		};
	    }
	};
    }
    // PersistentTreeNode subclass
    private static class Node<T>
	extends PersistentTreeNode<Node<T>,T,T> {
	/** The hash code of a {@link java.util.Set} with the
	 *  contents of the tree rooted at this node. */
	final int setHashCode;
	/** Size of the tree rooted at this node. */
	final int size;
	Node(T key, Node<T> left, Node<T> right) {
	    super(key, left, right);
	    this.setHashCode =
		((key==null)?0:key.hashCode()) + // this entry
		((left==null)?0:left.setHashCode) + // hash of left tree
		((right==null)?0:right.setHashCode); // hash of right tree
	    this.size = 1 + // this entry
		((left==null)?0:left.size) + // size of left tree
		((right==null)?0:right.size); // size of right tree
	}
	public T getValue() { return key; }
	static class Allocator<T>
	    extends PersistentTreeNode.Allocator<Node<T>,T,T> {
	    Node<T> newNode(T key, T value, Node<T> left, Node<T> right) {
		// ignore value.
		return new Node<T>(key, left, right);
	    }
	}
    }
}
